# -*- coding: utf-8 -*-

##
# This file was copied from Fabric-1.8.0 with some modifications.
#
# This distribution of fabric is distributed under the following BSD license:
#
#  Copyright (c) 2009, Christian Vest Hansen and Jeffrey E. Forcier
#  All rights reserved.
#
#  Redistribution and use in source and binary forms, with or without
#  modification, are permitted provided that the following conditions are met:
#
#      * Redistributions of source code must retain the above copyright notice,
#        this list of conditions and the following disclaimer.
#      * Redistributions in binary form must reproduce the above copyright
#        notice, this list of conditions and the following disclaimer in the
#        documentation and/or other materials provided with the distribution.
#
#  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
#  AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
#  ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
#  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
#  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
#  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
#  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
#  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
#  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
#  POSSIBILITY OF SUCH DAMAGE.
#
##


"""
This module contains Fab's `main` method plus related subroutines.

`main` is executed as the command line ``fab`` program and takes care of
parsing options and commands, loading the user settings file, loading a
fabfile, and executing the commands given.

The other callables defined in this module are internal only. Anything useful
to individuals leveraging Fabric as a library, should be kept elsewhere.
"""
import copy
import getpass
import logging
from operator import isMappingType
from optparse import Values, SUPPRESS_HELP
import os
import sys
import textwrap
import types

# For checking callables against the API, & easy mocking
from fabric import api, state
from fabric.contrib import console, files, project

from fabric.state import env_options
from fabric.tasks import Task, execute
from fabric.task_utils import _Dict, crawl
from fabric.utils import abort, indent, warn, _pty_size

from trinoadmin.util.exception import ConfigurationError, is_arguments_error
from trinoadmin import __version__
from trinoadmin.util.application import entry_point
from trinoadmin.util.fabric_application import FabricApplication
from trinoadmin.util.hiddenoptgroup import HiddenOptionGroup
from trinoadmin.util.parser import LoggingOptionParser

# One-time calculation of "all internal callables" to avoid doing this on every
# check of a given fabfile callable (in is_classic_task()).
_modules = [api, project, files, console]
_internals = reduce(lambda x, y: x + filter(callable, vars(y).values()), _modules, [])
_LOGGER = logging.getLogger(__name__)


def _get_presto_env_options():
    new_env_options = copy.deepcopy(env_options)
    commands_to_remove = ['fabfile', 'parallel', 'rcfile', 'skip_bad_hosts',
                          'warn_only', 'always_use_pty', 'skip_unknown_tasks',
                          'abort_on_prompts', 'pool_size',
                          'eagerly_disconnect', 'ssh_config_path']
    commands_to_hide = ['--roles', '--shell', '--linewise', '--show', '--hide',
                        '--gss-auth', '--gss-deleg', '--gss-kex', '--sudo-password']
    new_env_options = \
        [x for x in new_env_options if x.dest not in commands_to_remove]
    for env_option in new_env_options:
        if env_option.get_opt_string() in commands_to_hide:
            env_option.help = SUPPRESS_HELP
    return new_env_options


presto_env_options = _get_presto_env_options()


# Module recursion cache
class _ModuleCache(object):
    """
    Set-like object operating on modules and storing __name__s internally.
    """

    def __init__(self):
        self.cache = set()

    def __contains__(self, value):
        return value.__name__ in self.cache

    def add(self, value):
        return self.cache.add(value.__name__)

    def clear(self):
        return self.cache.clear()


_seen = _ModuleCache()


def is_classic_task(tup):
    """
    Takes (name, object) tuple, returns True if it's a non-Fab public callable.
    """
    name, func = tup
    try:
        is_classic = (
            callable(func) and (func not in _internals) and not
            name.startswith('_')
        )
    # Handle poorly behaved __eq__ implementations
    except (ValueError, TypeError):
        is_classic = False
    return is_classic


def load_fabfile(path, importer=None):
    """
    Import given fabfile path and return (docstring, callables).

    Specifically, the fabfile's ``__doc__`` attribute (a string) and a
    dictionary of ``{'name': callable}`` containing all callables which pass
    the "is a Fabric task" test.
    """
    if importer is None:
        importer = __import__
    # Get directory and fabfile name
    directory, fabfile = os.path.split(path)
    # If the directory isn't in the PYTHONPATH, add it so our import will work
    added_to_path = False
    index = None
    if directory not in sys.path:
        sys.path.insert(0, directory)
        added_to_path = True
    # If the directory IS in the PYTHONPATH, move it to the front temporarily,
    # otherwise other fabfiles -- like Fabric's own -- may scoop the intended
    # one.
    else:
        i = sys.path.index(directory)
        if i != 0:
            # Store index for later restoration
            index = i
            # Add to front, then remove from original position
            sys.path.insert(0, directory)
            del sys.path[i + 1]
    # Perform the import (trimming off the .py)
    imported = importer(os.path.splitext(fabfile)[0])
    # Remove directory from path if we added it ourselves (just to be neat)
    if added_to_path:
        del sys.path[0]
    # Put back in original index if we moved it
    if index is not None:
        sys.path.insert(index + 1, directory)
        del sys.path[0]

    # Actually load tasks
    docstring, new_style, classic, default = load_tasks_from_module(imported)
    tasks = new_style if state.env.new_style_tasks else classic
    # Clean up after ourselves
    _seen.clear()
    return docstring, tasks


def load_tasks_from_module(imported):
    """
    Handles loading all of the tasks for a given `imported` module
    """
    # Obey the use of <module>.__all__ if it is present
    imported_vars = vars(imported)
    if "__all__" in imported_vars:
        imported_vars = [(name, imported_vars[name]) for name in
                         imported_vars if name in imported_vars["__all__"]]
    else:
        imported_vars = imported_vars.items()
    # Return a two-tuple value.  First is the documentation, second is a
    # dictionary of callables only (and don't include Fab operations or
    # underscored callables)
    new_style, classic, default = extract_tasks(imported_vars)
    return imported.__doc__, new_style, classic, default


def extract_tasks(imported_vars):
    """
    Handle extracting tasks from a given list of variables
    """
    new_style_tasks = _Dict()
    classic_tasks = {}
    default_task = None
    if 'new_style_tasks' not in state.env:
        state.env.new_style_tasks = False
    for tup in imported_vars:
        name, obj = tup
        if is_task_object(obj):
            state.env.new_style_tasks = True
            # Use instance.name if defined
            if obj.name and obj.name != 'undefined':
                new_style_tasks[obj.name] = obj
            else:
                obj.name = name
                new_style_tasks[name] = obj
            # Handle aliasing
            if obj.aliases is not None:
                for alias in obj.aliases:
                    new_style_tasks[alias] = obj
            # Handle defaults
            if obj.is_default:
                default_task = obj
        elif is_classic_task(tup):
            classic_tasks[name] = obj
        elif is_task_module(obj):
            docs, newstyle, classic, default = load_tasks_from_module(obj)
            for task_name, task in newstyle.items():
                if name not in new_style_tasks:
                    new_style_tasks[name] = _Dict()
                new_style_tasks[name][task_name] = task
            if default is not None:
                new_style_tasks[name].default = default
    return new_style_tasks, classic_tasks, default_task


def is_task_module(a):
    """
    Determine if the provided value is a task module
    """
    # return (type(a) is types.ModuleType and
    #        any(map(is_task_object, vars(a).values())))
    if isinstance(a, types.ModuleType) and a not in _seen:
        # Flag module as seen
        _seen.add(a)
        # Signal that we need to check it out
        return True


def is_task_object(a):
    """
    Determine if the provided value is a ``Task`` object.

    This returning True signals that all tasks within the fabfile
    module must be Task objects.
    """
    return isinstance(a, Task) and a.use_task_objects


def parser_for_options():
    """
    Handle command-line options with LoggingOptionParser.

    Return parser, largely for use in `parse_arguments`.

    On this parser, you must call parser.parse_args()
    """
    #
    # Initialize
    #
    parser = LoggingOptionParser(
        usage='trino-admin [options] <command> [arg]',
        version='trino-admin %s' % __version__,
        epilog='\n' + '\n'.join(list_commands(None, 'normal')))

    #
    # Define options that don't become `env` vars (typically ones which cause
    # Fabric to do something other than its normal execution, such as
    # --version)
    #

    # Display info about a specific command
    parser.add_option(
        '-d',
        '--display',
        dest='display',
        action='store_true',
        default=False,
        help='print detailed information about command'
    )

    parser.add_option(
        '--extended-help',
        action='store_true',
        dest='extended_help',
        default=False,
        help='print out all options, including advanced ones'
    )

    parser.add_option(
        '-I',
        '--initial-password-prompt',
        action='store_true',
        default=False,
        help="Force password prompt up-front"
    )

    parser.add_option(
        '--nodeps',
        action='store_true',
        dest='nodeps',
        default=False,
        help=SUPPRESS_HELP
    )

    parser.add_option(
        '--no-config-update',
        action='store_true',
        dest='no_config_update',
        default=False,
        help="Don't update the cluster after rpm install"
    )

    parser.add_option(
        '--force',
        action='store_true',
        dest='force',
        default=False,
        help=SUPPRESS_HELP
    )

    #
    # Add in options which are also destined to show up as `env` vars.
    #

    advanced_options = HiddenOptionGroup(parser, "Advanced Options",
                                         suppress_help=True)

    # Hide most of the options from the help text so it's simpler. Need to
    # document the other options, however.
    commands_to_show = ['password']

    for option in presto_env_options:
        if option.dest in commands_to_show:
            parser.add_option(option)
        else:
            advanced_options.add_option(option)

    advanced_options.add_option(
        '--serial',
        action='store_true',
        dest='serial',
        default=False,
        help="default to serial execution method"
    )

    # Allow setting of arbitrary env vars at runtime.
    advanced_options.add_option(
        '--set',
        metavar="KEY=VALUE,...",
        dest='env_settings',
        default="",
        help=SUPPRESS_HELP
    )

    parser.add_option_group(advanced_options)

    # Return parser
    return parser


def _is_task(name, value):
    """
    Is the object a task as opposed to e.g. a dict or int?
    """
    return is_classic_task((name, value)) or is_task_object(value)


def _sift_tasks(mapping):
    tasks, collections = [], []
    for name, value in mapping.iteritems():
        if _is_task(name, value):
            tasks.append(name)
        elif isMappingType(value):
            collections.append(name)
    tasks = sorted(tasks)
    collections = sorted(collections)
    return tasks, collections


def _task_names(mapping):
    """
    Flatten & sort task names in a breadth-first fashion.

    Tasks are always listed before submodules at the same level, but within
    those two groups, sorting is alphabetical.
    """
    tasks, collections = _sift_tasks(mapping)
    for collection in collections:
        module = mapping[collection]
        if hasattr(module, 'default'):
            tasks.append(collection)
        tasks.extend(map(lambda x: " ".join((collection, x)),
                     _task_names(module)))
    return tasks


def _print_docstring(docstrings, name):
    if not docstrings:
        return False
    docstring = crawl(name, state.commands).__doc__
    if isinstance(docstring, basestring):
        return docstring


def _normal_list(docstrings=True):
    result = []
    task_names = _task_names(state.commands)
    # Want separator between name, description to be straight col
    max_len = reduce(lambda a, b: max(a, len(b)), task_names, 0)
    sep = '  '
    trail = '...'
    max_width = _pty_size()[1] - 1 - len(trail)
    for name in task_names:
        docstring = _print_docstring(docstrings, name)
        if docstring:
            lines = filter(None, docstring.splitlines())
            first_line = lines[0].strip()
            # Truncate it if it's longer than N chars
            size = max_width - (max_len + len(sep) + len(trail))
            if len(first_line) > size:
                first_line = first_line[:size] + trail
            output = name.ljust(max_len) + sep + first_line
        # Or nothing (so just the name)
        else:
            output = name
        result.append(indent(output))
    return result


COMMANDS_HEADER = 'Commands:'


def list_commands(docstring, format_):
    """
    Print all found commands/tasks, then exit. Invoked with ``-l/--list.``

    If ``docstring`` is non-empty, it will be printed before the task list.

    ``format_`` should conform to the options specified in
    ``LIST_FORMAT_OPTIONS``, e.g. ``"short"``, ``"normal"``.
    """
    # Short-circuit with simple short output
    if format_ == "short":
        return _task_names(state.commands)
    # Otherwise, handle more verbose modes
    result = []
    # Docstring at top, if applicable
    if docstring:
        trailer = "\n" if not docstring.endswith("\n") else ""
        result.append(docstring + trailer)
    header = COMMANDS_HEADER
    result.append(header)
    c = _normal_list()
    result.extend(c)
    result.extend("\n")
    return result


def get_task_docstring(task):
    details = [
        textwrap.dedent(task.__doc__)
        if task.__doc__
        else 'No docstring provided']

    return '\n'.join(details)


def display_command(name, code=0):
    """
    Print command function's docstring, then exit. Invoked with -d/--display.
    """
    # Sanity check
    command = crawl(name, state.commands)
    name = name.replace(".", " ")
    if command is None:
        msg = "Task '%s' does not appear to exist. Valid task names:\n%s"
        abort(msg % (name, "\n".join(_normal_list(False))))
    # get the presented docstring if found
    task_details = get_task_docstring(command)

    if task_details:
        print("Displaying detailed information for task '%s':" % name)
        print('')
        print(indent(task_details, strip=True))
        print('')
    # Or print notice if not
    else:
        print("No detailed information available for task '%s':" % name)
    sys.exit(code)


def parse_arguments(arguments, commands):
    """
    Parse string list into list of tuples: command, args.

    commands is formatted like {'install' : {'server' : WrappedCallable,
    'cli' : WrappedCallable}, 'topology': {'show' : WrappedCallable}}

    Thus, since our arguments are separated by spaces, and is of the form
    ['install', 'server'], we iterate through the commands, progressively
    going deeper into the dict.  If we run out of elements in the dict,
    the rest of the tokens are arguments to the function. If we don't
    get down to the bottom-most level, the command is not valid. If
    at any point the next token is not in the possible_cmd map, the
    command is invalid.
    """

    possible_cmds = commands.copy()
    pos = 0

    while pos < len(arguments):
        if not isinstance(possible_cmds, dict):
            # the rest of are all arguments to the cmd
            break
        if arguments[pos] not in possible_cmds:
            invalid_command_error(arguments)
        possible_cmds = possible_cmds[arguments[pos]]
        pos += 1

    if isinstance(possible_cmds, dict):
        invalid_command_error(arguments)

    cmds = [(".".join(arguments[:pos]), arguments[pos:], {}, [], [], [])]
    return cmds


def invalid_command_error(arguments):
    raise NameError("Command not found:\n%s" % indent(" ".join(arguments)))


def update_output_levels(show, hide):
    """
    Update state.output values as per given comma-separated list of key names.

    For example, ``update_output_levels(show='debug,warnings')`` is
    functionally equivalent to ``state.output['debug'] = True ;
    state.output['warnings'] = True``. Conversely, anything given to ``hide``
    sets the values to ``False``.
    """
    if show:
        for key in show.split(','):
            state.output[key] = True
    if hide:
        for key in hide.split(','):
            state.output[key] = False


def show_commands(docstring, format, code=0):
    print("\n".join(list_commands(docstring, format)))
    sys.exit(code)


def run_tasks(task_list):
    for name, args, kwargs, arg_hosts, arg_roles, arg_excl_hosts in task_list:
        try:
            nodeps_tasks = ['package.install', 'server.uninstall',
                            'server.install', 'server.upgrade']
            if state.env.nodeps and name.strip() not in nodeps_tasks:
                sys.stderr.write('Invalid argument --nodeps to task: %s\n'
                                 % name)
                display_command(name, 2)

            return execute(
                name,
                hosts=state.env.hosts,
                roles=arg_roles,
                exclude_hosts=state.env.exclude_hosts,
                *args, **kwargs
            )
        except TypeError as e:
            if is_arguments_error(e):
                print("Incorrect number of arguments to task.\n")
                _LOGGER.error('Incorrect number of arguments to task',
                              exc_info=True)
                display_command(name, 2)
            else:
                raise
        except BaseException as e:
            raise


def _escape_split(sep, argstr):
    """
    Allows for escaping of the separator: e.g. task:arg='foo\, bar'

    It should be noted that the way bash et. al. do command line parsing, those
    single quotes are required.
    """
    escaped_sep = r'\%s' % sep

    if escaped_sep not in argstr:
        return argstr.split(sep)

    before, _, after = argstr.partition(escaped_sep)
    startlist = before.split(sep)  # a regular split is fine here
    unfinished = startlist[-1]
    startlist = startlist[:-1]

    # recurse because there may be more escaped separators
    endlist = _escape_split(sep, after)

    # finish building the escaped value. we use endlist[0] becaue the first
    # part of the string sent in recursion is the rest of the escaped value.
    unfinished += sep + endlist[0]

    return startlist + [unfinished] + endlist[1:]  # put together all the parts


def _to_boolean(string):
    """
    Parses the given string into a boolean.  If its already a boolean, its
    returned unchanged.

    This method does strict parsing; only the string "True" returns the boolean
    True, and only the string "False" returns the boolean False.  All other
    values throw a ValueError.

    Args:
        string: the string to parse
    """
    if string is True or string == 'True':
        return True
    elif string is False or string == 'False':
        return False

    raise ValueError("invalid boolean string: %s" % string)


def _handle_generic_set_env_vars(non_default_options):
    if not hasattr(non_default_options, 'env_settings'):
        return non_default_options

    # Allow setting of arbitrary env keys.
    # This comes *before* the "specific" env_options so that those may
    # override these ones. Specific should override generic, if somebody
    # was silly enough to specify the same key in both places.
    # E.g. "fab --set shell=foo --shell=bar" should have env.shell set to
    # 'bar', not 'foo'.
    for pair in _escape_split(',', non_default_options.env_settings):
        pair = _escape_split('=', pair)
        # "--set x" => set env.x to True
        # "--set x=" => set env.x to ""
        key = pair[0]
        value = True
        if len(pair) == 2:
            try:
                value = _to_boolean(pair[1])
            except ValueError:
                value = pair[1]
        state.env[key] = value

    non_default_options_dict = vars(non_default_options)
    del non_default_options_dict['env_settings']
    return Values(non_default_options_dict)


def validate_hosts(cli_hosts, config_path):
    # If there's no config file to validate against, don't. This would happen
    # in the case of a task that doesn't define a callback that loads config.
    if config_path is None:
        return

    # At this point, state.env.conf_hosts contains the hosts that we loaded
    # from the configuration, if any.
    cli_host_set = set(cli_hosts.split(','))
    if 'conf_hosts' in state.env:
        conf_hosts = set(state.env.conf_hosts)
        if not cli_host_set.issubset(conf_hosts):
            raise ConfigurationError('Hosts defined in --hosts/-H must be '
                                     'present in %s' % (config_path))
    else:
        raise ConfigurationError(
            'Hosts cannot be defined with --hosts/-H when no hosts are listed '
            'in the configuration file %s. Correct the configuration file or '
            'run the command again without the --hosts or -H option.' %
            config_path)


def _update_env(default_options, non_default_options, load_config_callback):
    # Fill in the state with the default values
    for opt, value in default_options.__dict__.items():
        state.env[opt] = value

    if load_config_callback:
        config_path = load_config(load_config_callback)
    else:
        config_path = None

    # Save env.hosts from the config into another env variable for validation.
    # _handle_generic_set_env_vars will overwrite it if --set hosts=...
    # is present.
    if state.env.hosts:
        state.env.conf_hosts = state.env.hosts

    non_default_options = _handle_generic_set_env_vars(non_default_options)

    if isinstance(state.env.hosts, basestring):
        # Take advantage of the fact that if there was a generic --set option
        # for hosts, it's still an unsplit, comma separated string rather than
        # a list, which is what it would be after loading hosts from a config
        # file.
        validate_hosts(state.env.hosts, config_path)

    # Go back through and add the non-default values (e.g. the values that
    # were set on the CLI)
    for opt, value in non_default_options.__dict__.items():
        # raise error if hosts not in topology
        if opt == 'hosts':
            validate_hosts(value, config_path)

        state.env[opt] = value

    # Handle --hosts, --roles, --exclude-hosts (comma separated string =>
    # list)
    for key in ['hosts', 'roles', 'exclude_hosts']:
        if key in state.env and isinstance(state.env[key], basestring):
            state.env[key] = state.env[key].split(',')

    state.output['running'] = False
    state.output['status'] = False
    update_output_levels(show=state.env.show, hide=state.env.hide)
    state.env.skip_bad_hosts = True

    # env.conf_hosts is an implementation detail of the option parsing and
    # validation. Hide it from the world.
    if 'conf_hosts' in state.env:
        del state.env['conf_hosts']


def get_default_options(options, non_default_options):
    """
    Given a dictionary of options containing the defaults optparse has filled
    in, and a dictionary of options containing only options parsed from the
    command line, returns a dictionary containing the default options that
    remain after removing the default options that were overridden by the
    options passed on the command line.

    Mathematically, this returns a dictionary with
    default_options.keys = options.keys() \ non_default_options.keys()
    where \ is the set difference operator.
    The value of a key present in default_options is the value of the same key
    in options.
    """
    options_dict = vars(options)
    non_default_options_dict = vars(non_default_options)
    default_options = Values(dict((k, options_dict[k]) for k in options_dict
                                  if k not in non_default_options_dict))
    return default_options


def _get_config_callback(commands_to_run):
    config_callback = None
    if len(commands_to_run) != 1:
        raise Exception('Multiple commands are not supported')

    c = commands_to_run[0][0]
    module, command = c.split('.')

    module_dict = state.commands[module]
    command_callable = module_dict[command]

    try:
        config_callback = command_callable.pa_config_callback
    except AttributeError:
        pass

    return config_callback


def parse_and_validate_commands(args=sys.argv[1:]):
    # Find local fabfile path or abort
    fabfile = "trinoadmin"

    # Store absolute path to fabfile in case anyone needs it
    state.env.real_fabfile = fabfile

    # Load fabfile (which calls its module-level code, including
    # tweaks to env values) and put its commands in the shared commands
    # dict
    docstring, callables = load_fabfile(fabfile)
    state.commands.update(callables)

    # Parse command line options
    parser = parser_for_options()

    # Unless you pass in values, optparse fills in the default values for all
    # of the options. We want to save the version of the options without
    # default values, because that takes precedence over all other env vars.
    non_default_options, arguments = parser.parse_args(args, values=Values())
    options, arguments = parser.parse_args(args)
    default_options = get_default_options(options, non_default_options)

    # Handle regular args vs -- args
    arguments = parser.largs

    if len(parser.rargs) > 0:
        warn("Arbitrary remote shell commands not supported.")
        show_commands(None, 'normal', 2)

    if options.extended_help:
        parser.print_extended_help()
        sys.exit(0)

    # If user didn't specify any commands to run, show help
    if not arguments:
        parser.print_help()
        sys.exit(0)  # don't consider this an error

    # Parse arguments into commands to run (plus args/kwargs/hosts)
    commands_to_run = None
    try:
        commands_to_run = parse_arguments(arguments, state.commands)
    except NameError as e:
        warn(e.message)
        _LOGGER.warn("Unable to parse arguments", exc_info=True)
        parser.print_help()
        sys.exit(2)

    # Handle show (command-specific help) option
    if options.display:
        display_command(commands_to_run[0][0])

    load_config_callback = _get_config_callback(commands_to_run)
    _update_env(default_options, non_default_options, load_config_callback)

    if not options.serial:
        state.env.parallel = True

    state.env.warn_only = False

    # Initial password prompt, if requested
    if options.initial_password_prompt:
        prompt = "Initial value for env.password: "
        state.env.password = getpass.getpass(prompt)

    state.env['tasks'] = [x[0] for x in commands_to_run]

    return commands_to_run


def load_config(load_config_callback):
    """
    This provides a patch point for the unit tests so that individual test
    cases don't need to know the internal details of what happens in
    _load_topology. See test_main.py for examples.
    """
    return load_config_callback()


def _exit_code(results):
    """
    results from run_tasks take the form of a dict with one or more entries
    hostname: Exception | None

    If every entry in the dict has a value of None, the exit code is 0.
    If any entry has a value that is not None, something failed, and we should
    exit with a non-zero exit code.

    That isn't really the whole story: Any task that calls tasks.execute() and
    returns that as the result will have an item in the dictionary of the form
    hostname: {hostname: Exception | None}. This means that we need to
    recursively check any values in the map that are of type dict following the
    above scheme.
    """
    for v in results.values():
        # No exception, inspect the next value.
        if v is None:
            continue

        # The value is a dict resulting from calling fabric.tasks.execute.
        # Check the results recursively
        if type(v) is dict:
            exit_code = _exit_code(v)
            if exit_code != 0:
                return exit_code
            continue

        # In any case where things were OK above, we've continued the loop. At
        # this point, we know something failed.
        return 1
    return 0


@entry_point('trino-admin', version=__version__,
             log_file_path="trino-admin.log",
             application_class=FabricApplication)
def main(args=sys.argv[1:]):
    """
    Main command-line execution loop.
    """
    commands_to_run = parse_and_validate_commands(args)

    names = ", ".join(x[0] for x in commands_to_run)
    _LOGGER.info("Commands to run: %s" % names)

    # At this point all commands must exist, so execute them in order.
    return _exit_code(run_tasks(commands_to_run))


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Module for parsing and processing semantic versions
"""
class SemanticVersion(object):
    def __init__(self, version):
        self.version = version
        version_fields = self.version.split('.')
        if len(version_fields) > 3:
            exit('Version %s has more than 3 fields' % self.version)
        self.major_version = self._get_version_field_value(version_fields, 0)
        self.minor_version = self._get_version_field_value(version_fields, 1)
        self.patch_version = self._get_version_field_value(version_fields, 2)

    def _get_version_field_value(self, version_fields, index):
        try:
            return int(version_fields[index])
        except IndexError:
            # The field value was omitted for the version
            return 0
        except ValueError:
            exit('Version %s has a non-numeric field' % self.version)

    def __lt__(self, other):
        if self.major_version == other.major_version:
            if self.minor_version == other.minor_version:
                return self.patch_version < other.patch_version
            else:
                return self.minor_version < other.minor_version
        else:
            return self.major_version < other.major_version

    def __eq__(self, other):
        return self.major_version == other.major_version and \
               self.minor_version == other.minor_version and \
               self.patch_version == other.patch_veresion

    def __str__(self):
        return self.version

    @staticmethod
    def _bump_version(version_field):
        return str(int(version_field) + 1)

    def _get_acceptable_major_version_bumps(self):
        acceptable_major = self._bump_version(self.major_version)
        return [acceptable_major,
                acceptable_major + '.0',
                acceptable_major + '.0.0']

    def _get_acceptable_minor_version_bumps(self):
        acceptable_minor = self._bump_version(self.minor_version)
        return [str(self.major_version) + '.' + acceptable_minor,
                str(self.major_version) + '.' + acceptable_minor + '.0']

    def _get_acceptable_patch_version_bumps(self):
        acceptable_patch = self._bump_version(self.patch_version)
        return [str(self.major_version) + '.' + str(self.minor_version) + '.' + acceptable_patch]

    def get_acceptable_version_bumps(self):
        """
        This functions takes as input strings major, minor, and patch which should be
        the corresponding semvar fields for a release. It returns a list of strings, which
        contains all acceptable versions. For each field bump, lower fields may be omitted
        or 0s. For instance, bumping 0.1.2's major version can result in 1, 1.0, or 1.0.0.
        """
        major_bumps = self._get_acceptable_major_version_bumps()
        minor_bumps = self._get_acceptable_minor_version_bumps()
        patch_bumps = self._get_acceptable_patch_version_bumps()
        return major_bumps + minor_bumps + patch_bumps



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""Monkey patches needed to change logging and error handling in Fabric"""
import traceback
import sys
import logging
from traceback import format_exc

from fabric import state
from fabric.context_managers import settings
from fabric.exceptions import NetworkError
from fabric.job_queue import JobQueue
from fabric.tasks import _is_task, WrappedCallableTask, requires_parallel
from fabric.task_utils import crawl, parse_kwargs
from fabric.utils import error
import fabric.api
import fabric.operations
import fabric.tasks
from fabric.network import needs_host, to_dict, disconnect_all

from trinoadmin.util import exception


_LOGGER = logging.getLogger(__name__)
old_warn = fabric.utils.warn
old_abort = fabric.utils.abort
old_run = fabric.operations.run
old_sudo = fabric.operations.sudo


# Need to monkey patch Fabric's warn method in order to print out
# all exceptions seen to the logs.
def warn(msg):
    if fabric.api.env.host:
        msg = '[' + fabric.api.env.host + '] ' + msg
    old_warn(msg)
    _LOGGER.warn(msg + '\n\n' + format_exc())

fabric.utils.warn = warn
fabric.api.warn = warn


def abort(msg):
    if fabric.api.env.host:
        msg = '[' + fabric.api.env.host + '] ' + msg
    old_abort(msg)

fabric.utils.abort = abort
fabric.api.abort = abort


# Monkey patch run and sudo so that the stdout and stderr
# also go to the logs.
@needs_host
def run(command, shell=True, pty=True, combine_stderr=None, quiet=False,
        warn_only=False, stdout=None, stderr=None, timeout=None,
        shell_escape=None):
    out = old_run(command, shell=shell, pty=pty,
                  combine_stderr=combine_stderr, quiet=quiet,
                  warn_only=warn_only, stdout=stdout, stderr=stderr,
                  timeout=timeout, shell_escape=shell_escape)
    log_output(out)
    return out


fabric.operations.run = run
fabric.api.run = run


@needs_host
def sudo(command, shell=True, pty=True, combine_stderr=None, user=None,
         quiet=False, warn_only=False, stdout=None, stderr=None, group=None,
         timeout=None, shell_escape=None):
    out = old_sudo(command, shell=shell, pty=pty,
                   combine_stderr=combine_stderr, user=user, quiet=quiet,
                   warn_only=warn_only, stdout=stdout, stderr=stderr,
                   group=group, timeout=timeout, shell_escape=shell_escape)
    log_output(out)
    return out


fabric.operations.sudo = sudo
fabric.api.sudo = sudo


def log_output(out):
    return
    _LOGGER.info('\nCOMMAND: ' + out.command + '\nFULL COMMAND: ' +
                 out.real_command + '\nSTDOUT: ' + out + '\nSTDERR: ' +
                 out.stderr)


# Monkey patch _execute and execute so that we can handle errors differently
def _execute(task, host, my_env, args, kwargs, jobs, queue, multiprocessing):
    """
    Primary single-host work body of execute().
    """
    # Log to stdout
    if state.output.running and not hasattr(task, 'return_value'):
        print("[%s] Executing task '%s'" % (host, my_env['command']))
    # Create per-run env with connection settings
    local_env = to_dict(host)
    local_env.update(my_env)
    # Set a few more env flags for parallelism
    if queue is not None:
        local_env.update({'parallel': True, 'linewise': True})
    # Handle parallel execution
    if queue is not None:  # Since queue is only set for parallel
        name = local_env['host_string']

        # Wrap in another callable that:
        # * expands the env it's given to ensure parallel, linewise, etc are
        # all set correctly and explicitly. Such changes are naturally
        # insulted from the parent process.
        # * nukes the connection cache to prevent shared-access problems
        # * knows how to send the tasks' return value back over a Queue
        # * captures exceptions raised by the task
        def inner(args, kwargs, queue, name, env):
            state.env.update(env)

            def submit(result):
                queue.put({'name': name, 'result': result})

            try:
                state.connections.clear()
                submit(task.run(*args, **kwargs))
            except BaseException, e:
                _LOGGER.error(traceback.format_exc())
                submit(e)
                sys.exit(1)

        # Stuff into Process wrapper
        kwarg_dict = {
            'args': args,
            'kwargs': kwargs,
            'queue': queue,
            'name': name,
            'env': local_env,
        }
        p = multiprocessing.Process(target=inner, kwargs=kwarg_dict)
        # Name/id is host string
        p.name = name
        # Add to queue
        jobs.append(p)
    # Handle serial execution
    else:
        with settings(**local_env):
            return task.run(*args, **kwargs)


def execute(task, *args, **kwargs):
    """
    Patched version of fabric's execute task with alternative error handling
    """
    my_env = {'clean_revert': True}
    results = {}
    # Obtain task
    is_callable = callable(task)
    if not (is_callable or _is_task(task)):
        # Assume string, set env.command to it
        my_env['command'] = task
        task = crawl(task, state.commands)
        if task is None:
            msg = "%r is not callable or a valid task name" % (
                my_env['command'],)
            if state.env.get('skip_unknown_tasks', False):
                warn(msg)
                return
            else:
                abort(msg)
    # Set env.command if we were given a real function or callable task obj
    else:
        dunder_name = getattr(task, '__name__', None)
        my_env['command'] = getattr(task, 'name', dunder_name)
    # Normalize to Task instance if we ended up with a regular callable
    if not _is_task(task):
        task = WrappedCallableTask(task)
    # Filter out hosts/roles kwargs
    new_kwargs, hosts, roles, exclude_hosts = parse_kwargs(kwargs)
    # Set up host list
    my_env['all_hosts'], my_env[
        'effective_roles'] = task.get_hosts_and_effective_roles(hosts, roles,
                                                                exclude_hosts,
                                                                state.env)

    parallel = requires_parallel(task)
    if parallel:
        # Import multiprocessing if needed, erroring out usefully
        # if it can't.
        try:
            import multiprocessing
        except ImportError:
            import traceback

            tb = traceback.format_exc()
            abort(tb + """
    At least one task needs to be run in parallel, but the
    multiprocessing module cannot be imported (see above
    traceback.) Please make sure the module is installed
    or that the above ImportError is fixed.""")
    else:
        multiprocessing = None

    # Get pool size for this task
    pool_size = task.get_pool_size(my_env['all_hosts'], state.env.pool_size)
    # Set up job queue in case parallel is needed
    queue = multiprocessing.Queue() if parallel else None
    jobs = JobQueue(pool_size, queue)
    if state.output.debug:
        jobs._debug = True

    # Call on host list
    if my_env['all_hosts']:
        # Attempt to cycle on hosts, skipping if needed
        for host in my_env['all_hosts']:
            try:
                results[host] = _execute(
                    task, host, my_env, args, new_kwargs, jobs, queue,
                    multiprocessing
                )
            except NetworkError, e:
                results[host] = e
                # Backwards compat test re: whether to use an exception or
                # abort
                if state.env.skip_bad_hosts or state.env.warn_only:
                    func = warn
                else:
                    func = abort
                error(e.message, func=func, exception=e.wrapped)
            except SystemExit, e:
                results[host] = e

            # If requested, clear out connections here and not just at the end.
            if state.env.eagerly_disconnect:
                disconnect_all()

        # If running in parallel, block until job queue is emptied
        if jobs:
            jobs.close()
            # Abort if any children did not exit cleanly (fail-fast).
            # This prevents Fabric from continuing on to any other tasks.
            # Otherwise, pull in results from the child run.
            ran_jobs = jobs.run()
            for name, d in ran_jobs.iteritems():
                if d['exit_code'] != 0:
                    if isinstance(d['results'], NetworkError):
                        func = warn if state.env.skip_bad_hosts \
                            or state.env.warn_only else abort
                        error(d['results'].message,
                              exception=d['results'].wrapped, func=func)
                    elif exception.is_arguments_error(d['results']):
                        raise d['results']
                    elif isinstance(d['results'], SystemExit):
                        # System exit indicates abort
                        pass
                    elif isinstance(d['results'], BaseException):
                        error(d['results'].message, exception=d['results'])
                    else:
                        error('One or more hosts failed while executing task.')
                results[name] = d['results']

    # Or just run once for local-only
    else:
        with settings(**my_env):
            results['<local-only>'] = task.run(*args, **new_kwargs)
    # Return what we can from the inner task executions

    return results


fabric.tasks._execute = _execute
fabric.tasks.execute = execute



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This modules contains read-only constants used throughout
the presto admin project.
"""
from __future__ import print_function

import os
import sys

import trinoadmin

# Logging Config File Locations
LOGGING_CONFIG_FILE_NAME = 'trino-admin-logging.ini'
LOGGING_CONFIG_FILE_DIRECTORIES = [
    os.path.join(trinoadmin.main_dir, 'trinoadmin')
]

# Use the BRAND variable to define whether facebook's presto or trinodb's trino is being used at the time
# with the default value being the former.
# You can override the default by defining the environment variable PRESTO_TYPE.
# For example, PRESTO_TYPE=trino means use the latter
BRAND = os.environ.get('PRESTO_TYPE', 'trino').lower()
if BRAND == 'trinodb':
    BRAND = 'trino'
if BRAND not in ['trino', 'presto']:
    print("PRESTO_TYPE only support trino and presto")
    BRAND = 'trino'

# local configuration
LOG_DIR_ENV_VARIABLE = 'PRESTO_ADMIN_LOG_DIR'
CONFIG_DIR_ENV_VARIABLE = 'PRESTO_ADMIN_CONFIG_DIR'
LOCAL_CONF_DIR = '.trinoadmin'
DEFAULT_LOCAL_CONF_DIR = os.path.join(os.path.expanduser('~'), LOCAL_CONF_DIR)
TOPOLOGY_CONFIG_FILE = 'config.json'
COORDINATOR_DIR_NAME = 'coordinator'
WORKERS_DIR_NAME = 'workers'
CATALOG_DIR_NAME = 'catalog'

# remote configuration
REMOTE_CONF_DIR = '/etc/' + BRAND
REMOTE_CATALOG_DIR = os.path.join(REMOTE_CONF_DIR, 'catalog')
REMOTE_PACKAGES_PATH = '/opt/trinoadmin/packages'
DEFAULT_PRESTO_SERVER_LOG_FILE = '/var/log/' + BRAND + '/server.log'
DEFAULT_PRESTO_LAUNCHER_LOG_FILE = '/var/log/' + BRAND + '/launcher.log'
REMOTE_PLUGIN_DIR = '/usr/lib/' + BRAND + '/plugin'
REMOTE_COPY_DIR = '/tmp'

# Presto configuration files
CONFIG_PROPERTIES = "config.properties"
LOG_PROPERTIES = "log.properties"
JVM_CONFIG = "jvm.config"
NODE_PROPERTIES = "node.properties"


# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module to add extensions and helpers for fabric api methods
"""

from functools import wraps

from fabric.api import env, put, settings, sudo
from fabric.utils import abort


def get_host_list():
    return [host for host in env.hosts if host not in env.exclude_hosts]


def get_coordinator_role():
    return env.roledefs['coordinator']


def get_worker_role():
    return env.roledefs['worker']


def task_by_rolename(rolename):
    def inner_decorator(f):
        @wraps(f)
        def wrapper(*args, **kwargs):
            return by_rolename(env.host, rolename, f, *args, **kwargs)
        return wrapper
    return inner_decorator


def by_rolename(host, rolename, f, *args, **kwargs):
    if rolename is None:
        f(*args, **kwargs)
    else:
        if rolename not in env.roledefs.keys():
            abort("Invalid role name %s. Valid rolenames are %s" %
                  (rolename, env.roledefs.keys()))
        if host in env.roledefs[rolename]:
            return f(*args, **kwargs)


def by_role_coordinator(host, f, *args, **kwargs):
    if host in get_coordinator_role():
        return f(*args, **kwargs)


def by_role_worker(host, f, *args, **kwargs):
    if host in get_worker_role() and host not in get_coordinator_role():
        return f(*args, **kwargs)


def put_secure(user_group, mode, *args, **kwargs):
    missing_owner_code = 42
    user, group = user_group.split(":")

    files = put(*args, mode=mode, **kwargs)

    for pfile in files:
        with settings(warn_only=True):
            command = \
                "( getent passwd {user} >/dev/null || ( rm -f {file} ; " \
                "exit {missing_owner_code} ) ) && " \
                "chown {user_group} {file}".format(
                    user=user, file=pfile, user_group=user_group,
                    missing_owner_code=missing_owner_code)

            result = sudo(command)

            if result.return_code == missing_owner_code:
                abort("User %s does not exist. Make sure the Presto "
                      "server RPM is installed and try again" % (user,))
            elif result.failed:
                abort("Failed to chown file %s" % (pfile,))



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for setting and validating the presto-admin config
"""
import pprint

from fabric.api import env, runs_once, task

from trinoadmin.standalone.config import StandaloneConfig
from trinoadmin.util.base_config import requires_config

import trinoadmin.util.fabricapi as util


@task
@runs_once
@requires_config(StandaloneConfig)
def show():
    """
    Shows the current topology configuration for the cluster (including the
    coordinators, workers, SSH port, and SSH username)
    """
    pprint.pprint(get_conf_from_fabric(), width=1)


def get_conf_from_fabric():
    return {'coordinator': util.get_coordinator_role()[0],
            'workers': util.get_worker_role(),
            'port': env.port,
            'username': env.user}



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module defines error types relevant to the Presto
administrative suite.
"""
import re

import sys
import traceback


# Beware the nuances of pickling Exceptions:
# http://bugs.python.org/issue1692335
class ExceptionWithCause(Exception):

    def __init__(self, message=''):
        self.inner_exception = None

        causing_exception = sys.exc_info()[1]
        if causing_exception:
            self.inner_exception = traceback.format_exc() + \
                ExceptionWithCause.get_cause_if_supported(causing_exception)

        super(ExceptionWithCause, self).__init__(message)

    @staticmethod
    def get_cause_if_supported(exception):
        try:
            inner = exception.inner_exception
        except AttributeError:
            inner = None

        if inner:
            return '\nCaused by:\n{tb}'.format(
                tb=inner
            )
        else:
            return ''


class InvalidArgumentError(ExceptionWithCause):
    pass


class ConfigurationError(ExceptionWithCause):
    pass


class ConfigFileNotFoundError(ConfigurationError):
    def __init__(self, message='', config_path=''):
        super(ConfigFileNotFoundError, self).__init__(message)
        self.config_path = config_path


def is_arguments_error(exception):
    return isinstance(exception, TypeError) and \
        re.match(r'.+\(\) takes (at most \d+|no|exactly \d+|at least \d+) '
                 r'arguments? \(\d+ given\)', exception.message)



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for setting and validating the presto-admin config
"""
import os
import re

from fabric.api import env
from overrides import overrides

import trinoadmin.util.fabricapi as util
from trinoadmin import config
from trinoadmin.util import constants
from trinoadmin.trinoclient import CERTIFICATE_ALIAS
from trinoadmin.util.base_config import BaseConfig, SingleConfigItem
from trinoadmin.util.exception import ConfigurationError
from trinoadmin.util.local_config_util import get_topology_path
from trinoadmin.util.validators import validate_username, validate_port, \
    validate_host

# Created by the presto-server RPM package.
PRESTO_STANDALONE_USER = os.environ.get('PRESTO_USER', constants.BRAND)
PRESTO_STANDALONE_GROUP = os.environ.get('PRESTO_GROUP', constants.BRAND)
PRESTO_STANDALONE_USER_GROUP = "%s:%s" % (PRESTO_STANDALONE_USER,
                                          PRESTO_STANDALONE_GROUP)

# CONFIG KEYS
USERNAME = 'username'
PORT = 'port'
COORDINATOR = 'coordinator'
WORKERS = 'workers'

STANDALONE_CONFIG_LOADED = 'standalone_config_loaded'

PRESTO_ADMIN_PROPERTIES = ['username', 'port', 'coordinator', 'workers',
                           'java_home', CERTIFICATE_ALIAS]

DEFAULT_PROPERTIES = {USERNAME: 'root',
                      PORT: 22,
                      COORDINATOR: 'localhost',
                      WORKERS: ['localhost']}


def validate_coordinator(coordinator):
    validate_host(coordinator)
    return coordinator


def validate_workers_for_prompt(workers):
    return validate_workers(workers.split())


_TOPOLOGY_CONFIG = [
    SingleConfigItem(
        USERNAME, 'Enter user name for SSH connection to all nodes:',
        default=DEFAULT_PROPERTIES[USERNAME], validate=validate_username),
    SingleConfigItem(
        PORT, 'Enter port number for SSH connections to all nodes:',
        default=DEFAULT_PROPERTIES['port'], validate=validate_port),
    SingleConfigItem(
        COORDINATOR,
        'Enter host name or IP address for coordinator node. '
        'Enter an external host name or ip address if this is a multi-node '
        'cluster:',
        default=DEFAULT_PROPERTIES['coordinator'],
        validate=validate_coordinator),
    SingleConfigItem(
        WORKERS,
        'Enter host names or IP addresses for worker nodes separated by spaces:',
        default=' '.join(DEFAULT_PROPERTIES['workers']),
        validate=validate_workers_for_prompt)
]


def validate_java_home(java_home):
    return java_home


def validate(conf):
    for key in conf.keys():
        if key not in PRESTO_ADMIN_PROPERTIES:
            raise ConfigurationError('Invalid property: ' + key)

    try:
        username = conf['username']
    except KeyError:
        pass
    else:
        conf['username'] = validate_username(username)

    try:
        java_home = conf['java_home']
    except KeyError:
        pass
    else:
        conf['java_home'] = validate_java_home(java_home)

    try:
        coordinator = conf['coordinator']
    except KeyError:
        pass
    else:
        conf['coordinator'] = validate_coordinator(coordinator)

    try:
        workers = conf['workers']
    except KeyError:
        pass
    else:
        workers = [h for host in workers for h in _expand_host(host)]
        conf['workers'] = validate_workers(workers)

    try:
        port = conf['port']
    except KeyError:
        pass
    else:
        conf['port'] = validate_port(port)
    return conf


def validate_workers(workers):
    if not isinstance(workers, list):
        raise ConfigurationError('Workers must be of type list.  Found ' +
                                 str(type(workers)) + '.')

    if len(workers) < 1:
        raise ConfigurationError('Must specify at least one worker')

    for worker in workers:
        validate_host(worker)
    return workers


def _expand_host(host):
    match = re.match("(.*)\[(\d{1,})-(\d{1,})\](.*)", host)
    if match is not None and len(match.groups()) == 4:
        prefix, start, end, suffix = match.groups()
        if int(start) > int(end):
            raise ValueError("the range must be in ascending order")
        if len(start) == len(end) and len(start) > 1:
            number_format = "{0:0" + str(len(start)) + "d}"
            host_list = [number_format.format(i) for i in range(int(start), int(end) + 1)]
            return [_format_hostname(prefix, i, suffix) for i in host_list]
        else:
            return [_format_hostname(prefix, i, suffix) for i in range(int(start), int(end) + 1)]
    else:
        return [host]


def _format_hostname(prefix, number, suffix):
    return "{prefix}{num}{suffix}".format(prefix=prefix, num=number, suffix=suffix)


class StandaloneConfig(BaseConfig):
    def __init__(self):
        super(StandaloneConfig, self).__init__(get_topology_path(), _TOPOLOGY_CONFIG)

    @overrides
    def read_conf(self):
        conf = self._get_conf_from_file()
        config.fill_defaults(conf, DEFAULT_PROPERTIES)
        validate(conf)
        return conf

    def _get_conf_from_file(self):
        return config.get_conf_from_json_file(self.config_path)

    @overrides
    def is_config_loaded(self):
        return STANDALONE_CONFIG_LOADED in env and env[STANDALONE_CONFIG_LOADED]

    @overrides
    def set_config_loaded(self):
        env[STANDALONE_CONFIG_LOADED] = True

    @overrides
    def set_env_from_conf(self, conf):
        self.config = conf
        env.user = conf['username']
        env.port = conf['port']
        try:
            env.java_home = conf['java_home']
        except KeyError:
            env.java_home = None
        env.roledefs['coordinator'] = [conf['coordinator']]
        env.roledefs['worker'] = conf['workers']
        env.roledefs['all'] = self._dedup_list(util.get_coordinator_role() +
                                               util.get_worker_role())

        env.hosts = env.roledefs['all'][:]
        env.conf = conf

    @staticmethod
    def _dedup_list(host_list):
        deduped_list = []
        for item in host_list:
            if item not in deduped_list:
                deduped_list.append(item)
        return deduped_list



# -*- coding: utf-8 -*-

#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import logging
from fabric.context_managers import settings, hide
from fabric.operations import sudo
from fabric.tasks import execute
from trinoadmin.util.exception import ConfigurationError
from trinoadmin.util.constants import DEFAULT_PRESTO_LAUNCHER_LOG_FILE,\
    DEFAULT_PRESTO_SERVER_LOG_FILE, REMOTE_CONF_DIR, REMOTE_CATALOG_DIR
import trinoadmin.util.validators

_LOGGER = logging.getLogger(__name__)

NODE_CONFIG_FILE = REMOTE_CONF_DIR + '/node.properties'
GENERAL_CONFIG_FILE = REMOTE_CONF_DIR + '/config.properties'


def lookup_port(host):
    """
    Get the http port from config.properties http-server.http.port property
    if available.
    If the property is missing return default port 8080.
    If the file is missing or cannot parse the port number,
    throw ConfigurationError
    :param host:
    :return:
    """
    port = lookup_in_config('http-server.http.port', GENERAL_CONFIG_FILE, host)
    if not port:
        _LOGGER.info('Could not find property http-server.http.port.'
                     'Defaulting to 8080.')
        return 8080
    try:
        port = port.split('=', 1)[1]
        port = trinoadmin.util.validators.validate_port(port)
        _LOGGER.info('Looked up port ' + str(port) + ' on host ' +
                     host)
        return port
    except ConfigurationError as e:
        raise ConfigurationError(e.message +
                                 ' for property '
                                 'http-server.http.port on host ' +
                                 host + '.')


def lookup_server_log_file(host):
    try:
        return lookup_string_config('node.server-log-file', NODE_CONFIG_FILE,
                                    host, DEFAULT_PRESTO_SERVER_LOG_FILE)
    except:
        return DEFAULT_PRESTO_SERVER_LOG_FILE


def lookup_launcher_log_file(host):
    try:
        return lookup_string_config('node.launcher-log-file', NODE_CONFIG_FILE,
                                    host, DEFAULT_PRESTO_LAUNCHER_LOG_FILE)
    except:
        return DEFAULT_PRESTO_LAUNCHER_LOG_FILE


def lookup_catalog_directory(host):
    try:
        return lookup_string_config('catalog.config-dir', NODE_CONFIG_FILE,
                                    host, REMOTE_CATALOG_DIR)
    except:
        return REMOTE_CATALOG_DIR


def lookup_string_config(config_value, config_file, host, default=''):
    value = lookup_in_config(config_value, config_file, host)
    if value:
        return value.split('=', 1)[1]
    else:
        return default


def lookup_in_config(config_key, config_file, host):
    with settings(hide('stdout', 'warnings', 'aborts')):
        config_value = execute(sudo, 'grep %s= %s' % (config_key, config_file),
                               user='presto',
                               warn_only=True, host=host)[host]

    if isinstance(config_value, Exception) or config_value.return_code == 2:
        raise ConfigurationError('Could not access config file %s on '
                                 'host %s' % (config_file, host))

    return config_value



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for the presto coordinator's configuration.
Loads and validates the coordinator.json file and creates the files needed
to deploy on the presto cluster
"""
import copy
import logging

from fabric.api import env

from trinoadmin.util import constants
from trinoadmin.node import Node
from trinoadmin.trino_conf import validate_presto_conf
from trinoadmin.util.exception import ConfigurationError
from trinoadmin.util.local_config_util import get_coordinator_directory

_LOGGER = logging.getLogger(__name__)


class Coordinator(Node):
    DEFAULT_PROPERTIES = {'node.properties':
                          {'node.environment': constants.BRAND,
                           'node.data-dir': '/var/lib/{}/data'.format(constants.BRAND),
                           'node.launcher-log-file':
                               '/var/log/{}/launcher.log'.format(constants.BRAND),
                           'node.server-log-file':
                               '/var/log/{}/server.log'.format(constants.BRAND),
                           'catalog.config-dir': '/etc/{}/catalog'.format(constants.BRAND),
                           'plugin.dir': '/usr/lib/{}/lib/plugin'.format(constants.BRAND)},
                          'jvm.config': ['-server',
                                         '-Xmx16G',
                                         '-XX:-UseBiasedLocking',
                                         '-XX:+UseG1GC',
                                         '-XX:G1HeapRegionSize=32M',
                                         '-XX:+ExplicitGCInvokesConcurrent',
                                         '-XX:+ExitOnOutOfMemoryError',
                                         '-XX:+UseGCOverheadLimit',
                                         '-XX:+HeapDumpOnOutOfMemoryError',
                                         '-XX:ReservedCodeCacheSize=512M',
                                         '-Djdk.attach.allowAttachSelf=true',
                                         '-Djdk.nio.maxCachedBufferSize=2000000',
                                         '-DHADOOP_USER_NAME=hive',  # not Presto default
                                         ],
                          'config.properties': {
                              'coordinator': 'true',
                              'discovery-server.enabled': 'true',
                              'http-server.http.port': '8080',
                              'node-scheduler.include-coordinator': 'false',
                              'query.max-memory': '50GB',
                              'query.max-memory-per-node': '8GB'}
                          }

    def _get_conf_dir(self):
        return get_coordinator_directory()

    def default_config(self, filename):
        try:
            conf = copy.deepcopy(self.DEFAULT_PROPERTIES[filename])
        except KeyError:
            raise ConfigurationError('Invalid configuration file name: %s' %
                                     filename)
        if filename == 'config.properties':
            coordinator = env.roledefs['coordinator'][0]
            workers = env.roledefs['worker']
            if coordinator in workers:
                conf['node-scheduler.include-coordinator'] = 'true'
            conf['discovery.uri'] = 'http://%s:8080' % coordinator
        return conf

    @staticmethod
    def validate(conf):
        validate_presto_conf(conf)
        if 'coordinator' not in conf['config.properties']:
            raise ConfigurationError('Must specify coordinator=true in '
                                     'coordinator\'s config.properties')
        if conf['config.properties']['coordinator'] != 'true':
            raise ConfigurationError('Coordinator cannot be false in the '
                                     'coordinator\'s config.properties.')
        return conf



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module for common configuration stuff.
"""

import abc

from functools import wraps

from fabric.context_managers import settings
from fabric.operations import prompt

from trinoadmin import config
from trinoadmin.config import ConfigFileNotFoundError
from trinoadmin.util.exception import ConfigurationError


class SingleConfigItem(object):
    def __init__(self, key, prompt, default=None, validate=None):
        self.key = key
        self.prompt = prompt
        self.default = default
        self.validate = validate

    def prompt_user(self, conf):
        conf[self.key] = prompt(self.prompt,
                                default=conf.get(self.key, self.default),
                                validate=self.validate)

    def collect_prompts(self, l):
        l.append((self.prompt, self.key))


class MultiConfigItem(object):
    def __init__(self, items, validate, validate_keys,
                 validate_failed_text):
        self.items = items
        self.validate = validate
        self.validate_keys = validate_keys
        self.validate_failed_text = validate_failed_text

    def prompt_user(self, conf):
        while True:
            for item in self.items:
                item.prompt_user(conf)

            validate_args = [conf[k] for k in self.validate_keys]
            if self.validate(*validate_args):
                break
            print (self.validate_failed_text % self.validate_keys) % conf

    def collect_prompts(self, l):
        for item in self.items:
            item.collect_prompts(l)


def requires_config(config_class):
    def wrap(func):
        config_instance = config_class()
        func.pa_config_callback = config_instance.get_config

        @wraps(func)
        def wrapper(*args, **kwargs):
            if not config_instance.is_config_loaded():
                raise ConfigurationError('Required config not loaded at task '
                                         'execution time.')
            return func(*args, **kwargs)
        return wrapper
    return wrap


class BaseConfig(object):
    '''
    BaseConfig provides the common config functionality for loading
    configuration files for presto-admin and going through the interactive
    config process if a config file isn't present.

    Instances of classes that subclass BaseConfig are intended to be used with
    the @requires_config decorator, which is responsible for adding an
    attribute to the task that tells main() how to load the configuration
    and subsequently for enforcing that the configuration has been loaded at
    the time the task is actually run.

    In order to be compatible with @requires_config, subclasses must define
    a no-arguments constructor.
    '''
    __metaclass__ = abc.ABCMeta

    def __init__(self, config_path, config_items):
        self.config_path = config_path
        self.config_items = config_items
        self.config = {}

    def __getitem__(self, key):
        return self.config[key]

    def __setitem__(self, key, value):
        self.config[key] = value

    def __delitem__(self, key):
        del self.config[key]

    def read_conf(self):
        return config.get_conf_from_json_file(self.config_path)

    def write_conf(self, conf):
        config.write(config.json_to_string(conf), self.config_path)
        return self.config_path

    def get_conf_interactive(self):
        conf = {}
        for item in self.config_items:
            item.prompt_user(conf)
        return conf

    def get_config(self):
        with settings(parallel=False):
            if not self.is_config_loaded():
                conf = {}
                try:
                    conf = self.read_conf()
                except ConfigFileNotFoundError:
                    conf = self.get_conf_interactive()
                    self.write_conf(conf)

                self.set_env_from_conf(conf)
                self.set_config_loaded()
            return self.config_path

    @abc.abstractmethod
    def is_config_loaded(self):
        pass

    @abc.abstractmethod
    def set_config_loaded(self):
        pass

    @abc.abstractmethod
    def set_env_from_conf(self, conf):
        pass



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Module for reading, writing, and processing configuration files
"""
import json
import os
import logging
import errno
import re

from trinoadmin.util.exception import ConfigurationError,\
    ConfigFileNotFoundError

COMMENT_CHARS = ['!', '#']
_LOGGER = logging.getLogger(__name__)


def get_conf_from_json_file(path):
    try:
        with open(path, 'r') as conf_file:
            if os.path.getsize(conf_file.name) == 0:
                return {}
            return json.load(conf_file)
    except IOError:
        raise ConfigFileNotFoundError(
            config_path=path, message="Missing configuration file %s." %
            (repr(path)))
    except ValueError as e:
        raise ConfigurationError(e)


def get_conf_from_properties_file(path):
    with open(path, 'r') as conf_file:
        return get_conf_from_properties_data(conf_file)


def get_conf_from_properties_data(data):
    props = {}
    for line in data.read().splitlines():
        line = line.strip()
        if len(line) > 0 and line[0] not in COMMENT_CHARS:
            pair = split_to_pair(line)
            props[pair[0]] = pair[1]
    return props


def split_to_pair(line):
    split_line = re.split(r'\s*(?<!\\):\s*|\s*(?<!\\)=\s*|(?<!\\)\s+', line,
                          maxsplit=1)
    if len(split_line) != 2:
        raise ConfigurationError(
            line + " is not in the expected format: <property>=<value>, "
                   "<property>:<value> or <property> <value>")
    return tuple(split_line)


def get_conf_from_config_file(path):
    with open(path, 'r') as conf_file:
        settings = conf_file.read().splitlines()
        return settings


def json_to_string(conf):
    return json.dumps(conf, indent=4, separators=(',', ':'))


def write_conf_to_file(conf, path):
    # Note: this function expects conf to be flat
    # either a dict for .properties file or a list for .config
    ext = os.path.splitext(path)[1]
    if ext == ".properties":
        write_properties_file(conf, path)
    elif ext == ".config":
        write_config_file(conf, path)


def write_properties_file(conf, path):
    output = ''
    for key, value in conf.iteritems():
        output += '%s=%s\n' % (key, value)
    write(output, path)


def write_config_file(conf, path):
    output = '\n'.join(conf)
    write(output, path)


def write(output, path):
    conf_directory = os.path.dirname(path)
    try:
        os.makedirs(conf_directory)
    except OSError as e:
        if e.errno == errno.EEXIST:
            pass
        else:
            raise

    with open(path, 'w') as f:
        f.write(output)


def fill_defaults(conf, defaults):
    try:
        default_items = defaults.iteritems()
    except AttributeError:
        return

    for k, v in default_items:
        conf.setdefault(k, v)
        fill_defaults(conf[k], v)



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Logic for starting and stopping Fabric applications.
"""

from fabric.network import disconnect_all
from trinoadmin.util.application import Application

import logging
import sys


# Normally this would use the logger for __name__, however, this is
# effectively the "root" logger for the application.  If this code
# were running directly in the executable script __name__ would be
# set to '__main__', so we emulate that same behavior here.  This should
# resolve to the same logger that will be used by the entry point script.
logger = logging.getLogger('__main__')


class FabricApplication(Application):
    """
    A Presto Fabric application entry point.  Provides logging and exception
    handling features.  Additionally cleans up Fabric network connections
    before exiting.
    """

    def _exit_cleanup_hook(self):
        """
        Disconnect all Fabric connections in addition to shutting down the
        logging.
        """
        disconnect_all()
        Application._exit_cleanup_hook(self)

    def _handle_error(self):
        """
        Handle KeyboardInterrupt in a special way: don't indicate
        that it's an error.

        Returns:
            Nothing
        """
        self._log_exception()
        if isinstance(self.exception, KeyboardInterrupt):
            print >> sys.stderr, "Stopped."
            sys.exit(0)
        else:
            Application._handle_error(self)



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
An extension to optparse for presto-admin which logs user parsing errors.
"""

import logging
from optparse import OptionParser
import sys


_LOGGER = logging.getLogger(__name__)


class LoggingOptionParser(OptionParser):
    """
    An extension to optparse which logs exceptions via the logging
    module in addition to writing the out to stderr.

    If used with HiddenOptionGroup, print_extended_help disables the
    suppress_help attribute of HiddenOptionGroup so as to print out
    extended helptext.
    """

    def exit(self, status=0, msg=None):
        _LOGGER.debug("Exiting option parser!")
        if msg:
            sys.stderr.write(msg)
            _LOGGER.error(msg)
        sys.exit(status)

    def print_extended_help(self, filename=None):
        old_suppress_help = {}
        for group in self.option_groups:
            try:
                old_suppress_help[group] = group.suppress_help
                group.suppress_help = False
            except AttributeError as e:
                old_suppress_help[group] = None
                _LOGGER.debug("Option group does not have option to "
                              "suppress help; exception is " + e.message)
        self.print_help(file=filename)

        for group in self.option_groups:
            # Restore the suppressed help when applicable
            if old_suppress_help[group]:
                group.suppress_help = True

    def format_epilog(self, formatter):
        """
        The default format_epilog strips the newlines (using textwrap),
        so we override format_epilog here to use its own epilog
        """
        if not self.epilog:
            self.epilog = ""
        return self.epilog



# -*- coding: utf-8 -*-
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Module for sending HTTP requests
"""
import urllib2


def send_get_request(url):
    response = None
    try:
        response = urllib2.urlopen(url)
        if response.getcode() != 200:
            exit('Get request to %s responded with status of %s' % (url, str(response.getcode())))
        else:
            headers = response.info()
            contents = response.read()
            return headers, contents
    finally:
        if response:
            response.close()


def send_authorized_post_request(url, data, authorization_string, content_type, content_length):
    response = None
    try:
        request = urllib2.Request(url, data,
                                  {'Content-Type': '%s' % content_type,
                                   'Content-Length': content_length,
                                   'Authorization': 'Basic %s' % authorization_string})
        response = urllib2.urlopen(request)
        status = response.getcode()
        headers = response.info()
        contents = response.read()
        if status != 201:
            print headers
            print contents
            exit('Failed to post to %s' % url)
    finally:
        if response:
            response.close()
